#
# gateway.py - Gateway base module.
# Copyright (c) 2005 shinGETsu Project.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: gateway.py,v 1.52 2006/01/19 12:34:34 fuktommy Exp $
#

import cgi
import os.path
import re
import urllib
import sys
import time

import basecgi
import config
from cache import *
from node import *

__version__ = "$Revision: 1.52 $"


class Message(dict):

    """Multi-language message for gateway."""

    def __init__(self, file):
        dict.__init__(self)
        try:
            f = open(file)
            del_eos = re.compile(r"[\r\n]*")
            iscomment = re.compile(r"^#$").search
            for line in f:
                line = del_eos.sub("", line)
                if iscomment(line):
                    pass
                else:
                    buf = line.split("<>")
                    if len(buf) == 2:
                        buf[1] = urllib.unquote(buf[1])
                        self[buf[0]] = buf[1]
            f.close()
        except IOError:
            sys.stderr.write(file + ": IOError\n")

# End of Message


def search_message(accept_language):
    """Search message file.

    Example of accept_language: "ja,en-us;q=0.7,en;q=0.3"
    """
    q = {}
    lang = []
    if accept_language != "":
        for i in accept_language.split(","):
            found = re.search(r"(\S+)\s*;\s*q=(\S+)", i)
            if found:
                try:
                    lang = found.group(1)
                    q[lang] = float(found.group(2))
                except ValueError:
                    pass
            else:
                q[i] = 1

        lang = q.keys()
        lang.sort(lambda a,b: cmp(q[b], q[a]))
    lang.append(config.language)
    for i in lang:
        file = config.file_dir + "/" + "message-" + i + ".txt"
        if os.path.isfile(file):
            return Message(file)
    return None

# End of search_message


class CGI(basecgi.CGI):

    root = config.root_path
    sep = config.query_separator
    appli = config.application
    gateway_cgi = config.gateway
    list_cgi = config.list_cgi
    thread_cgi = config.thread_cgi
    note_cgi = config.note_cgi
    admin_cgi = config.admin_cgi
    message = None

    def __init__(self,
                 stdin=sys.stdin,
                 stdout=sys.stdout,
                 stderr=sys.stderr,
                 environ=os.environ):
        basecgi.CGI.__init__(self,
                             stdin=stdin,
                             stdout=stdout,
                             stderr=stderr,
                             environ=environ)
        if "HTTP_ACCEPT_LANGUAGE" in self.environ:
            al = self.environ["HTTP_ACCEPT_LANGUAGE"]
        else:
            al = ""
        self.message = search_message(al)
        addr = self.environ.get("REMOTE_ADDR", "")
        self.isadmin = re.search(config.admin, addr)
        self.isfriend = re.search(config.friend, addr)
        self.isvisitor = re.search(config.visitor, addr)

    def path_info(self):
        """Parse PATH_INFO.

        If PATH_INFO is not defined, use QUERY_STRING.
        x.cgi?foo&bar=y -> path="foo".
        """
        m = re.search(r"([^&;=]*)(&|$)", self.environ.get("QUERY_STRING", ""))
        if self.environ.get("PATH_INFO", "") != "":
            path = self.environ["PATH_INFO"]
            if path.startswith("/"):
                path = path[1:]
        elif m is not None:
            path = m.group(1)
        else:
            path = ""
        path = self.escape(self.str_decode(path))
        return path

    def str_encode(self, query):
        """Encode for URI.

        ~ -> %7E
        """
        return urllib.quote(query)

    def str_decode(self, query):
        """Decode URI.

        %7E -> ~
        """
        return urllib.unquote(query)

    def file_encode(self, type, query):
        """Encode for filename.

        ~ -> 7E
        """
        buf = [type, "_"]
        for i in query:
            buf.append("%02X" % ord(i))
        return "".join(buf)

    def file_decode(self, query):
        """Decode filename.

        7E -> ~
        """
        q = query.split("_")
        if q < 1:
            return None
        else:
            query = q[1]
        buf = []
        for i in range(0, len(query), 2):
            try:
                buf.append("%c" % int(query[i:i+2], 16))
            except (ValueError, IndexError):
                self.stderr.write(query + ": ValueError/IndexError\n")
                return None
        return "".join(buf)

    def escape(self, msg):
        msg = msg.replace("&", "&amp;")
        msg = re.sub(r"&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);",
                     r"&\1;",
                     msg)
        msg = msg.replace("<", "&lt;")
        msg = msg.replace(">", "&gt;")
        msg = msg.replace("\r", "")
        msg = msg.replace("\n", "<br>")
        return msg

    def menubar(self, id="", rss=""):
        if id:
            id = ' id="%s"' % id
        if rss == "":
            rss = self.gateway_cgi + self.sep + "rss"
        message = self.message
        buf = []
        self.stdout.write(
            '<p class="menu"%s>\n' % id +
            '  <a href="' + self.gateway_cgi + '">' +
            message["top"] + "</a> |\n" +
            '  <a href="' + self.list_cgi + '">' +
            message["menu"] + "</a> |\n")
        if self.isfriend or self.isadmin:
            buf.append('  <a href="%s%snew">%s</a>' %
                       (self.gateway_cgi, self.sep, message["new"]))
        if self.isadmin:
            for i in ("changes", "search"):
                buf.append('  <a href="%s%s%s">%s</a>' %
                           (self.admin_cgi, self.sep, i, message[i]))
        buf.append('  <a href="' + rss + '">' + message["rss"] + "</a>")
        self.stdout.write(" |\n".join(buf) + "\n</p>\n")

    def header(self, title="", rss="", onload="",
                     cookie=None, deny_robot=False):
        """Print CGI and HTTP header."""
        if rss == "":
            rss = self.gateway_cgi + self.sep + "rss"
        message = self.message
        self.stdout.write("Content-Type: text/html; charset=UTF-8\r\n")
        if cookie:
            self.stdout.write("%s\r\n" % cookie)
        self.stdout.write(
            "\r\n" +
            '<?xml version="1.0" encoding="UTF-8"?>\n' +
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"\n'+
            '  "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">\n' +
            '<html xmlns="http://www.w3.org/1999/xhtml"' +
            ' lang="%s" xml:lang="%s">\n' %
                (message["lang"], message["lang"]) +
            '<head>\n' +
            '  <meta http-equiv="content-type"' +
            ' content="text/html; charset=UTF-8" />\n' +
            '  <meta http-equiv="content-script-type"' +
            ' content="text/javascript" />\n' +
            "  <title>" + title + "</title>\n")
        if deny_robot:
            self.stdout.write('  <meta name="robots" content="NOINDEX" />\n')
        self.stdout.write(
            '  <link rev="made" href="http://shingetsu.info/" />\n'+
            '  <link rel="contents" href="' + self.list_cgi + '" />\n' +
            '  <link rel="stylesheet" type="text/css"' +
            ' href="' + config.css + '" />\n' +
            '  <link rel="alternate" type="application/rss+xml" title="RSS"' +
            ' href="' + rss + '" />\n' +
            "</head>\n")
        if onload:
            self.stdout.write('<body onload="%s">\n' % onload)
        else:
            self.stdout.write("<body>\n")
        self.menubar("top", rss=rss)
        if self.isadmin:
            self.stdout.write(
                '<h1><a href="%s/search%s%s" title="%s">' %
                (self.admin_cgi, self.sep,
                 self.str_encode(title), message["search"]) +
                '%s</a></h1>\n' % title)
        else:
            self.stdout.write('<h1>%s</h1>\n' % title)
        self.stdout.write(
            '<script type="text/javascript" src="/tools.js"></script>\n')

    def footer(self):
        self.stdout.write(
            '<address>Powered by ' +
            '<a href="http://shingetsu.info/">shinGETsu</a>.</address>\n')
        self.stdout.write("</body>\n</html>\n")

    def localtime(self, stamp=0):
        """Return YYYY-mm-dd HH:MM."""
        return time.strftime('%Y-%m-%d %H:%M', time.localtime(int(stamp)))

    def rfc822_time(self, stamp=0):
        """Return date and time in RFC822 format."""
        return time.strftime("%a, %d %b %Y %H:%M:%S GMT",
                             time.gmtime(int(stamp)))

    def res_anchor(self, id):
        return ('<a href="#r%s" onmouseover="popupAnchor(event, \'%s\')"' + \
                ' onmouseout="hidePopup()">') % (id, id)

    def html_format(self, plain):
        buf = plain.replace("<br>", "\n")
        buf = self.escape(buf)
        buf = re.sub(r"<br>",
                     "<br />\n    ",
                     buf)
        buf = re.sub(r"(&gt;&gt;)([0-9a-f]{8})",
                     self.res_anchor(r"\2") + r"\g<0></a>",
                     buf)
        buf = re.sub(r"https?://[^\x00-\x20\"'()<>\[\]\x7F-\xFF]{2,}",
                     r'<a href="\g<0>">\g<0></a>',
                     buf)
        buf = re.sub(r"\[\[<a.*?>(.*?)\]\]</a>",
                     r"[[\1]]",
                     buf)

        tmp = ""
        while buf:
            m = re.search(r"\[\[([^<>]+?)\]\]", buf)
            if m is not None:
                tmp += buf[:m.start()]
                tmp += self.bracket_link(m.group(1))
                buf = buf[m.end():]
            else:
                tmp += buf
                buf = ""
        return tmp

    def bracket_link(self, link):
        """Encode bracket string to link.

        See WikiWikiWeb.
        """
        m = re.search(r"^/(thread)/([^/]+)/([0-9a-f]{8})$", link)
        if m is not None:
            uri = self.thread_cgi + self.sep + \
                  self.str_encode(m.group(2)) + \
                  "#r" + m.group(3)
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^/(list|thread|note)/([^/]+)$", link)
        if m is not None:
            uri = self.appli[m.group(1)] + self.sep + \
                  self.str_encode(m.group(2))
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^([^/]+)/([0-9a-f]{8})$", link)
        if m is not None:
            uri = self.appli[self.appli_type] + self.sep + \
                  self.str_encode(m.group(1)) + \
                  "#r" + m.group(2)
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^([^/]+)$", link)
        if m is not None:
            uri = self.appli[self.appli_type] + self.sep + \
                  self.str_encode(m.group(1))
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        return "[[" + link + "]]"

    def remove_file_form(self, cache):
        filesize = str(cache.size*10/1024/1024/10.0)
        if self.isadmin:
            self.stdout.write(
                '<form method="get" action="' + self.admin_cgi + '"><p>\n' +
                '  <input type="submit" value="' + self.message["del_file"] + 
                '" tabindex="9" accesskey="d" />\n' +
                '  <input type="hidden" name="cmd" value="delete" />\n' +
                '  <input type="hidden" name="file" value="' +
                cache.datfile + '" />\n' +
                "  %s%s\n" % (filesize, self.message["mb"]) +
                "</p></form>\n")
        else:
            self.stdout.write("<p>%s%s</p>\n" % (filesize, self.message["mb"]))

    def print302(self, next):
        """Print CGI header (302 moved temporarily)."""
        self.stdout.write(
            "Content-Type: text/html; charset=ASCII\n" +
            "\n" +
            '<?xml version="1.0" encoding="ASCII"?>\n' +
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"\n' +
            '  "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n' +
            '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">\n' +
            "<head>\n" +
            "  <title>moved</title>\n" +
            '  <meta name="robots" content="NOINDEX" />\n' +
            '  <link rev="made" href="http://shingetsu.info/" />\n'+
            '  <link rel="contents" href="' + self.list_cgi + '" />\n' +
            "</head>\n" +
            "<body>\n" +
            '<p>Click and jump to <a href="' + next + '">' + next + '</a>' +
            '</p>\n' +
            '<script type="text/javascript">' +
            'window.location.href = "' + next + '";</script>\n' +
            "</body>\n</html>\n")

    def print403(self):
        """Print CGI header (403 forbidden)."""
        self.header(self.message["403"], deny_robot=True)
        self.stdout.write("<p>" + self.message["403_body"] + "</p>")
        self.footer()

    def print404(self, cache=None):
        """Print CGI header (404 not found)."""
        self.header(self.message["404"], deny_robot=True)
        self.stdout.write("<p>" + self.message["404_body"] + "</p>")
        if cache is not None:
            self.remove_file_form(cache)
        self.footer()

    def lock(self):
        if not os.path.isfile(config.search_lock):
            f = file(config.search_lock, "wb")
            f.close()
            return True
        elif os.path.getmtime(config.search_lock) + \
             config.search_timeout < time.time():
            f = file(config.search_lock, "wb")
            f.close()
            return True
        else:
            return False

    def unlock(self):
        try:
            os.remove(config.search_lock)
        except (OSError, IOError), (errorno, strerror):
            self.stderr.write("OSError/IOError: " + strerror + "\n")
            return False

    def get_cache(self, cache):
        """Search cache from network."""
        if not self.lock():
            return False
        self.unlock()
        return cache.search()

    def print_new_element_form(self, parent=None):
        if not (self.isadmin or self.isfriend):
            return
        if parent is not None:
            datfile = parent.datfile
            cginame = self.list_cgi
        else:
            datfile = ""
            cginame = self.gateway_cgi

        self.stdout.write(
            '<form method="post" action="' + cginame + '"><p>\n' +
            '  <input type="hidden" name="cmd" value="new" />\n' +
            '  <input type="hidden" name="file" value="' + datfile + '" />\n' +
            '  <input type="submit" value="' + self.message["new"] + '"' +
            ' name="submit" tabindex="1" accesskey="w" />\n')

        for type in config.types:
            self.stdout.write(
                '  <input type="radio" value="' + type + '" name="type"' +
                ' tabindex="2" accesskey="t" />' +
                self.message[type] + "\n")

        if parent is not None:
            self.stdout.write(
                '  <input type="checkbox" name="dopost" value="dopost"' +
                ' checked="checked" tabindex="2" accesskey="s" />' +
                self.message["send"] + "\n")

        self.stdout.write(
            "  <br />\n" +
            self.message["name"] + ": " +
            '<input name="link" size="19" value=""' +
            ' tabindex="3" accesskey="t" />\n' +
            "</p></form>\n")

    def error_time(self):
        from random import gauss
        return int(gauss(time.time(), config.time_error))

    def do_post(self, path, form):
        """Post article."""
        import base64
        try:
            attach = form["attach"]
        except KeyError:
            attach = None
            str_attach = ""

        if (attach is not None) and attach.file:
            if (len(attach.value) > config.file_limit*1024*1024):
                self.header(self.message["big_file"], deny_robot=True)
                self.footer()
                return
            str_attach = base64.encodestring(attach.value).replace("\n", "")
        guess_suffix = "txt"
        if (attach is not None) and attach.filename != "":
            found = re.search(r"\.([^.]+)$", attach.filename)
            if found:
                guess_suffix = found.group(1).lower()

        suffix = form.getfirst("suffix", "")
        if (suffix == "") or (suffix == "AUTO"):
            suffix = guess_suffix
        elif suffix.startswith("."):
            suffix = suffix[1:].lower()
        else:
            suffix = suffix.lower()
        suffix = re.sub(r"[^0-9A-Za-z]", "", suffix)

        if form.getfirst("error", "") != "":
            stamp = self.error_time()
        else:
            stamp = int(time.time())

        body = {}
        for key in ("base_stamp", "base_id", "name", "mail", "body"):
            value = form.getfirst(key, "")
            if value != "":
                body[key] = self.escape(value)
        if str_attach != "":
            body["attach"] = str_attach
            body["suffix"] = re.sub(r"[\r\n]", "", suffix)

        if not body:
            self.header(self.message["null_article"], deny_robot=True)
            self.footer()
            return

        cache = Cache(form.getfirst("file"))
        rec = Record(datfile=cache.datfile)
        passwd = form.getfirst("passwd", "")
        id = rec.build(stamp, body, passwd=passwd)
        cache.add_data([rec])

        if form.getfirst("dopost", "") != "":
            updatelist = UpdateList()
            updatelist.append(rec)
            updatelist.sync()
            nodelist = NodeList()
            nodelist.tell_update(cache, stamp=stamp, id=id)

        return id[:8]

    def do_new_element(self, path, form):
        """Make new list element."""
        stamp = self.error_time()

        if form.getfirst("link", "") == "":
            self.header(self.message["null_name"], deny_robot=True)
            self.footer()
            return
        elif re.search(r"/|\]\]", form.getfirst("link", "")):
            self.header(self.message["bad_name"], deny_robot=True)
            self.footer()
            return
        elif form.getfirst("type", "") == "":
            self.header(self.message["null_type"], deny_robot=True)
            self.footer()
            return

        cache = Cache(form.getfirst("file"))
        rec = Record(datfile=cache.datfile)
        body = {}
        for key in ("type", "link"):
            value = form.getfirst(key, "")
            body[key] = self.escape(value)
            id = rec.build(stamp, body)
        cache.add_data([rec])

        if form.getfirst("dopost", "") != "":
            updatelist = UpdateList()
            updatelist.append(rec)
            updatelist.sync()
            nodelist = NodeList()
            nodelist.tell_update(cache, stamp=stamp, id=id)

        return True

    def check_get_cache(self):
        agent = self.environ.get("HTTP_USER_AGENT", "")
        host = self.environ.get("HTTP_HOST", "")
        if self.isadmin:
            return True
        elif not self.isfriend:
            return False
        elif re.search(config.robot, agent):
            return False
        else:
            return True

    def check_visitor(self):
        return self.isadmin or self.isfriend or self.isvisitor

# End of CGI
